"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Utilities to work with FindOptions.
 */
var FindOptionsUtils = (function () {
    function FindOptionsUtils() {
    }
    /**
     * Checks if given object is really instance of FindOneOptions interface.
     */
    FindOptionsUtils.isFindOneOptions = function (obj) {
        var possibleOptions = obj;
        return possibleOptions &&
            (possibleOptions.where instanceof Object ||
                possibleOptions.join instanceof Object ||
                possibleOptions.order instanceof Object);
    };
    /**
     * Checks if given object is really instance of FindManyOptions interface.
     */
    FindOptionsUtils.isFindManyOptions = function (obj) {
        var possibleOptions = obj;
        return possibleOptions &&
            (possibleOptions.where instanceof Object ||
                possibleOptions.join instanceof Object ||
                possibleOptions.order instanceof Object ||
                typeof possibleOptions.skip === "number" ||
                typeof possibleOptions.take === "number");
    };
    /**
     * Checks if given object is really instance of FindOptions interface.
     */
    FindOptionsUtils.extractFindOneOptionsAlias = function (object) {
        if (this.isFindOneOptions(object) && object.join)
            return object.join.alias;
        return undefined;
    };
    /**
     * Checks if given object is really instance of FindOptions interface.
     */
    FindOptionsUtils.extractFindManyOptionsAlias = function (object) {
        if (this.isFindManyOptions(object) && object.join)
            return object.join.alias;
        return undefined;
    };
    /**
     * Applies give find one options to the given query builder.
     */
    FindOptionsUtils.applyFindOneOptionsOrConditionsToQueryBuilder = function (qb, options) {
        if (this.isFindOneOptions(options))
            return this.applyOptionsToQueryBuilder(qb, options);
        if (options)
            return this.applyConditions(qb, options);
        return qb;
    };
    /**
     * Applies give find many options to the given query builder.
     */
    FindOptionsUtils.applyFindManyOptionsOrConditionsToQueryBuilder = function (qb, options) {
        if (this.isFindManyOptions(options))
            return this.applyOptionsToQueryBuilder(qb, options);
        if (options)
            return this.applyConditions(qb, options);
        return qb;
    };
    /**
     * Applies give find options to the given query builder.
     */
    FindOptionsUtils.applyOptionsToQueryBuilder = function (qb, options) {
        // if options are not set then simply return query builder. This is made for simplicity of usage.
        if (!options || !this.isFindOneOptions(options))
            return qb;
        // apply all options from FindOptions
        if (options.where)
            this.applyConditions(qb, options.where);
        if (options.skip)
            qb.skip(options.skip);
        if (options.take)
            qb.take(options.take);
        if (options.order)
            Object.keys(options.order).forEach(function (key) {
                qb.addOrderBy(qb.alias + "." + key, options.order[key]);
            });
        if (options.join) {
            if (options.join.leftJoin)
                Object.keys(options.join.leftJoin).forEach(function (key) {
                    qb.leftJoin(options.join.leftJoin[key], key);
                });
            if (options.join.innerJoin)
                Object.keys(options.join.innerJoin).forEach(function (key) {
                    qb.innerJoin(options.join.innerJoin[key], key);
                });
            if (options.join.leftJoinAndSelect)
                Object.keys(options.join.leftJoinAndSelect).forEach(function (key) {
                    qb.leftJoinAndSelect(options.join.leftJoinAndSelect[key], key);
                });
            if (options.join.innerJoinAndSelect)
                Object.keys(options.join.innerJoinAndSelect).forEach(function (key) {
                    qb.innerJoinAndSelect(options.join.innerJoinAndSelect[key], key);
                });
        }
        return qb;
    };
    /**
     * Applies given simple conditions set to a given query builder.
     */
    FindOptionsUtils.applyConditions = function (qb, conditions) {
        Object.keys(conditions).forEach(function (key, index) {
            if (conditions[key] === null) {
                qb.andWhere(qb.alias + "." + key + " IS NULL");
            }
            else {
                var parameterName = "where_" + index;
                qb.andWhere(qb.alias + "." + key + "=:" + parameterName)
                    .setParameter(parameterName, conditions[key]);
            }
        });
        return qb;
    };
    return FindOptionsUtils;
}());
exports.FindOptionsUtils = FindOptionsUtils;

//# sourceMappingURL=FindOptionsUtils.js.map
